/*****************************************************************************
 *
 * calculator.c
 *
 * Copyright (c) Fabalabs Software GmbH, 2005
 *   oss@fabalabs.org
 *   Author: Friedrich Priewasser (friedrich.priewasser@fabalabs.org)
 *   Last Modified:   2005-01-17
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include "queue.h"
#include "calculator.h"

//Read and verify input
int readInput(char* str, Queue* infixQueue){
	int itemWasOperator=TRUE;
	int unaryMin=FALSE;
	int brackets=0;
	int pos=0;
	long val=0;
	int number;
	char c=str[0];
	ItemType kind=OPERATOR;
	Item* item;
	while(c){
		if(c>='0' && c<='9'){
			if(kind==OPERATOR){
				if(!itemWasOperator) return -11;
				itemWasOperator=FALSE;
				unaryMin=FALSE;
				kind=OPERAND;
			}
			number=c-'0';
			val=val*10+number;
		}else{
			//If the last item was an operator enqueu it in the infix queue
			if(kind==OPERAND){
				item=(Item*)malloc(sizeof(Item));
				item->Type=OPERAND;
				item->Val=val;
				queuePushHead(infixQueue, item);
				val=0;
				kind=OPERATOR;
			}
			
			//After an operator only an operand (checked before), one unary minus, a left bracket or a space is allowed:
			if(itemWasOperator && (c!='-' || unaryMin) && c!='(') return -12;
			//A left bracket is only allowed after an operand:
			else if(!itemWasOperator && c=='(') return -13;
			
			if(c=='+'){
				enqueueOperator(infixQueue, PLUS);
			}else if(c=='-' && !itemWasOperator){
				enqueueOperator(infixQueue, MINUS);
			}else if(c=='*'){
				enqueueOperator(infixQueue, TIMES);
			}else if(c=='/'){
				enqueueOperator(infixQueue, DIVIDE);
			}else if(c=='^'){
				enqueueOperator(infixQueue, POWER);
			}else if(c=='-' && itemWasOperator){
				enqueueOperator(infixQueue, UNARY_MINUS);
				unaryMin=TRUE;
			}else if(c=='('){
				enqueueOperator(infixQueue, LEFT_BRACKET);
				brackets++;
				unaryMin=FALSE;
			}else if(c==')'){
				enqueueOperator(infixQueue, RIGHT_BRACKET);
				brackets--;
				itemWasOperator=FALSE;
			}else{
				return -14;
			}
			
			if(c!=')') itemWasOperator=TRUE;
		}
				
		pos++;
		c=str[pos];
	}
	
	//Enqueue the last operand
	if(kind==OPERAND){
		item=(Item*)malloc(sizeof(Item));
		item->Type=OPERAND;
		item->Val=val;
		queuePushHead(infixQueue, item);
	}else if(str[pos-1]==')'){
	}else{
		return -15;
	}
	if(brackets!=0) return -16;
	
	return 0;
}

void enqueueOperator(Queue* q, operators op){
	Item* item=(Item*)malloc(sizeof(Item));
	item->Type=OPERATOR;
	item->Val=op;
	queuePushHead(q, item);
}


//Transform infix expression to postfix expression
void transform(Queue* infixQueue,Queue* postfixQueue){
	Item* item;
	Item* op;
	Queue* operatorStack=queueNew();
	while(queueGetLength(infixQueue)>0){
		item=queuePopTail(infixQueue);
		if(item->Type==OPERAND){
			queuePushHead(postfixQueue, item);
		}else{
			switch(item->Val){
				case PLUS: case MINUS: case TIMES: case DIVIDE: case POWER: case UNARY_MINUS:
					while(queueGetLength(operatorStack)>0){
						op=queuePopHead(operatorStack);
						if(op->Val<item->Val && 
								!(op->Val==PLUS && item->Val==MINUS) &&
								!(op->Val==TIMES && item->Val==DIVIDE)){
							queuePushHead(operatorStack,op);
							break;
						}else{
							queuePushHead(postfixQueue,op);
						}
					}
					queuePushHead(operatorStack,item);
					break;
				case LEFT_BRACKET:
					queuePushHead(operatorStack,item);
					break;
				case RIGHT_BRACKET:
					while(queueGetLength(operatorStack)>0){
						op=queuePopHead(operatorStack);
						if(op->Val==LEFT_BRACKET){
							break;
						}else{
							queuePushHead(postfixQueue, op);
						}
					}
					break;
			}						
		}
		
	}
	while(queueGetLength(operatorStack)>0){
		op=queuePopHead(operatorStack);
		queuePushHead(postfixQueue,op);
	}
	queueDestroy(operatorStack);
}

int calculate(Queue* postfixQueue, long* result){
	Item* item;
	Item* operand;
	long val1;
	long val2;
	long val3;
	Queue* operandStack=queueNew();
	while(queueGetLength(postfixQueue)>0){
		item=queuePopTail(postfixQueue);
		if(item->Type==OPERAND){
			queuePushHead(operandStack, item);
		}else{
			operand=queuePopHead(operandStack);
			val2=operand->Val;
			free(operand);
			if(item->Val!=UNARY_MINUS){
				operand=queuePopHead(operandStack);
				val1=operand->Val;
				free(operand);
			}
			switch(item->Val){
				case PLUS:
					val3=val1+val2;
					break;
				case MINUS:
					val3=val1-val2;
					break;
				case TIMES:
					val3=val1*val2;
					break;
				case DIVIDE:
					if (val2==0) return -21;
					val3=val1/val2;
					break;
				case POWER:
					if(val2<0){
						if(val1!=0){
							val1=1/val1;
							val2=-val2;
						}else{
							return -22;
						}
					}
					val3=1;
					while(val2>0){
						val3*=val1;
						val2--;
					}
					break;
				case UNARY_MINUS:
					val3=-val2;
					break;
			}
			item->Type=OPERAND;
			item->Val=val3;
			queuePushHead(operandStack,item);
		}
	}
	item=queuePopHead(operandStack);
	*result=item->Val;
	free(item);
	queueDestroy(operandStack);
	return 0;
}

int evaluate(char* input, long* result){
	Queue* infixQueue=queueNew();
	Queue* postfixQueue=queueNew();
	int status;
	status=readInput(input, infixQueue);
	if(status==0){
		transform(infixQueue, postfixQueue);
		status=calculate(postfixQueue, result);
	}
	queueDestroy(infixQueue);
	queueDestroy(postfixQueue);
	return status;
}
