/*****************************************************************************
 *
 * queue.c
 *
 * Copyright (c) Fabalabs Software GmbH, 2005
 *   oss@fabalabs.org
 *   Author: Friedrich Priewasser (friedrich.priewasser@fabalabs.org)
 *   Last Modified:   2005-01-17
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include "queue.h"
#include <stdlib.h>

Queue* queueNew(void){
	Queue* q=(Queue*)malloc(sizeof(Queue));
	q->head=NULL;
	q->tail=NULL;
	q->length=0;
}

void queueDestroy(Queue* queue){
	if(queue) free(queue);
}

int queueGetLength(Queue* queue){
	return queue->length;
}

void queuePushHead(Queue* queue, void* val){
	struct Element* elem=(struct Element*)malloc(sizeof(struct Element));
	elem->data=val;
	elem->next=queue->head;
	elem->prev=NULL;
	if(queue->head) queue->head->prev=elem;
	else queue->tail=elem;
	queue->head=elem;
	queue->length++;
}

void queuePushTail(Queue* queue, void* val){
	struct Element* elem=(struct Element*)malloc(sizeof(struct Element));
	
	elem->data=val;
	elem->next=NULL;
	elem->prev=queue->tail;
	if(queue->tail) queue->tail->next=elem;
	else queue->head=elem;
	queue->tail=elem;
	queue->length++;
}

void* queuePopHead(Queue* queue){
	void* val;
	struct Element* elem;
	if(queue->head==NULL) return 0;
	elem=queue->head;
	queue->head=queue->head->next;
	if(queue->head) queue->head->prev=NULL;
	val=elem->data;
	free(elem);
	queue->length--;
	return val;
}

void* queuePopTail(Queue* queue){
	void* val;
	struct Element* elem;
	if(queue->tail==NULL) return 0;
	elem=queue->tail;
	queue->tail=queue->tail->prev;
	if(queue->tail) queue->tail->next=NULL;
	val=elem->data;
	free(elem);
	queue->length--;
	return val;
}
